// Copyright YALI AMIT, January 15, 2001.



// __________________________________________________________________________
// Includes

#include <cstdio>
#include <iostream>
#include <fstream>

#include "includes.h"
#include "image.h"
#include "simpstrg.h"
#include "viewgrid.h"
#include "globals.h"
#include "pixel.h"
void find_reflector(const dgrid& x, const dgrid& y, dgrid& Q);
void calc_QR_decomp(dgrid& A, dgrid& sigma, dgrid& gamma);
void calc_Qtransb(const dgrid& A, const dgrid& gamma, 
		  const dgrid& b, dgrid& c);
void calc_Q(const dgrid& A, const dgrid& gamma, dgrid& Q);
void calc_R(const dgrid& A, const dgrid& sigma, dgrid& R);
void backward_solve(const dgrid&A, const dgrid& y, dgrid& x);
#define pint(a) ( (a>=0) ? ((int)(a)) : ((int)(a)-1) )
// __________________________________________________________________________
// Find the reflector Q that transfroms x to y

void find_reflector(const dgrid& x, const dgrid& y, dgrid& Q)
{
  int i;
  dgrid u = x - y;
  double unorm = 0;
  for (i = 0; i < u.dimx(); ++i)
    unorm += u(i) * u(i);
  unorm = sqrt(unorm);
  for (i = 0; i < u.dimx(); ++i)
    u(i) /= unorm;
  Q.init(u.dimx(), u.dimx());
  Q.clear();
  for (i = 0; i < Q.dimx(); ++i)
    Q(i, i) = 1;
  Q = Q - (u * u.transpose() + u * u.transpose());

} // find_reflector

// __________________________________________________________________________
// Calculate the QR decomposition of a matrix A.  This is the algorithm
// described on p. 153 of [FMC], modified to work when m > n.  It
// doesn't actually calculate Q and R, but stores all the data needed
// to easily create them in A, sigma, and gamma.

void calc_QR_decomp(dgrid& A, dgrid& sigma, dgrid& gamma)
{
  int i, j, k;

  // Some initializations
  int m = A.dimx(), n = A.dimy();
  sigma.init(n);
  gamma.init(n);

  // Calculate a reflector for each column of A (including the last
  // as it is necessary to do so when m > n).
  for (i = 0; i < n; ++i)
  {
    // Initialize u_i by setting it equal to the vector to reflect.
    dgrid u(m - i);
    for (j = 0; j < u.dimx(); ++j)
      u(j) = A(j + i, i);

    // Calculate siqma_i, the 2-norm of the vector to reflect.
    double sigmasquared = 0;
    for (j = 0; j < u.dimx(); ++j)
      sigmasquared += u(j) * u(j);
    sigma(i) = sqrt(sigmasquared);
    if (u(0) < 0)
      sigma(i) = -sigma(i);

    // Correct the first entry of u
    u(0) += sigma(i);

    // Calculate the square of the 2-norm of u and use it to calculate gamma.
    double unormsquared = sigmasquared - A(i, i) * A(i, i) + u(0) * u(0);
    gamma(i) = 2 / unormsquared;

    // Store u_k in A
    for (j = 0; j < u.dimx(); ++j)
      A(j + i, i) = u(j);

    // Multiply the relevant rows of A by the reflector.
    for (j = i + 1; j < n; ++j)
    {
      double tau = 0;
      for (k = 0; k < u.dimx(); ++k)
        tau += u(k) * A(i + k, j);
      tau *= gamma(i);
      for (k = 0; k < u.dimx(); ++k)
        A(i + k, j) -= tau * u(k);
    }
  }

} // calc_QR_decomp

// __________________________________________________________________________
// Calculate Q_transpose * b using A, gamma, and b without calculating Q.

void calc_Qtransb(const dgrid& A, const dgrid& gamma,
		  const dgrid& b, dgrid& c)
{
  int i, j;

  c = b;
  for (i = 0; i < A.dimy(); ++i) 
  {
    dgrid u(A.dimx() - i);
    for (j = 0; j < u.dimx(); ++j)
      u(j) = A(i + j, i);

    double tau = 0;
    for (j = 0; j < u.dimx(); ++j)
      tau += u(j) * c(i + j);
    tau *= gamma(i);

    for (j = 0; j < u.dimx(); ++j)
      c(i + j) -= tau * u(j);
  }

} // calc_Qtransb

// __________________________________________________________________________
// Recover Q from A and gamma.  In practice this isn't done.

void calc_Q(const dgrid& A, const dgrid& gamma, dgrid& Q)
{
  int i, j, k;

  // Initialize Q to be the identity matrix
  Q.init(A.dimx(), A.dimx());
  Q.clear();
  for (i = 0; i < Q.dimx(); ++i)
    Q(i, i) = 1;

  // Apply Q_n,...,Q_1 to Q.  This is algorithm 3.2.17 on p. 150 of [FMC].
  for (i = A.dimy() - 1; i >= 0; --i)
  {
    dgrid u(Q.dimx() - i);
    for (j = 0; j < u.dimx(); ++j)
      u(j) = A(i + j, i);
    for (j = i; j < Q.dimy(); ++j)
    {
      double tau = 0;
      for (k = 0; k < u.dimx(); ++k)
        tau += u(k) * Q(i + k, j);
      tau *= gamma(i);
      for (k = 0; k < u.dimx(); ++k)
        Q(i + k, j) -= tau * u(k);
    }
  }

} // calc_Q

// __________________________________________________________________________
// Recover R from A and sigma

void calc_R(const dgrid& A, const dgrid& sigma, dgrid& R)
{
  int i, j;

  R.init(A.dimx(), A.dimy());
  R.clear();
  for (i = 0; i < R.dimx(); ++i)
    for (j = i + 1; j < R.dimy(); ++j)
      R(i, j) = A(i, j);
  for (i = 0; i < R.dimy(); ++i)
    R(i, i) = -sigma(i);

} // calc_R

// __________________________________________________________________________
// Use backward substitution on the upper part of A to solve Ax = y.

void backward_solve(const dgrid&A, const dgrid& y, dgrid& x)
{
  int i, j;

  x.init(y.dimx());
  for (i = x.dimx() - 1; i >= 0; --i)
  {
    x(i) = y(i) / A(i, i);
    for (j = i + 1; j < x.dimx(); ++j)
      x(i) -= A(i, j) * x(j) / A(i, i);
  }

} // backward_solve

// __________________________________________________________________________
// Given A and b, with Ax = b, find the QR decomposition of A 
// and solve for x.

void solve_lsq(dgrid& A, const dgrid& b, dgrid& x)
{
  dgrid c, R, sigma, gamma;

  calc_QR_decomp(A, sigma, gamma);
  calc_R(A, sigma, R);
  dgrid R1;
  R.subgrid(R1, 0, 0, R.dimy(), R.dimy());
  calc_Qtransb(A, gamma, b, c);
  dgrid c1;
  c.subgrid(c1, 0, 0, R.dimy(), 1);
  backward_solve(R1, c1, x);

} // solve_lsq

// __________________________________________________________________________

image& image::subtract_lsq(const ucgrid& mask, int xlow, int ylow)
{
  dgrid A, sigma, gamma;
  dgrid b, c, x, R;
  int i, j;

  int hasmask = (mask.dimx() > 0);
  int asize = 0;
  if (!hasmask)
    asize = dimx() * dimy();
  else {
    for (i = 0; i < mask.dimx(); ++i)
      for (j = 0; j < mask.dimy(); ++j)
        asize += (mask(i, j) > 0);
  }

  A.init(asize, 3);
  b.init(asize);
  int row = 0;
  for (i = xlow; i < dimx(); ++i)
    for (j = ylow; j < dimy(); ++j) {
      if (!hasmask || mask.get(i - xlow, j - ylow)) {
        A(row, 0) = i;
        A(row, 1) = j;
        A(row, 2) = 1;
        b(row) = (*this)(i, j);
        ++row;
      }
    }

  calc_QR_decomp(A, sigma, gamma);
  calc_R(A, sigma, R);
  dgrid R1;
  R.subgrid(R1, 0, 0, R.dimy(), R.dimy());
  calc_Qtransb(A, gamma, b, c);
  dgrid c1;
  c.subgrid(c1, 0, 0, R.dimy(), 1);
  backward_solve(R1, c1, x);

  for (i = 0; i < dimx(); ++i)
    for (j = 0; j < dimy(); ++j) {
      int tmpval = int(128 + (*this)(i, j) - 
		       x(0) * i - x(1) * j - x(2) + 0.5);
      if (tmpval < 0) 
	(*this)(i, j) = 0;
      else if (tmpval > 255)
	(*this)(i, j) = 255;
      else
	(*this)(i, j) = tmpval;
    }

  return *this;

} // image::subtract_lsq

// __________________________________________________________________________

void subtract_lsq(dgrid& d, const ucgrid& mask, int xlow, int ylow)
{
  dgrid A, sigma, gamma;
  dgrid b, c, x, R;
  int i, j;

  int hasmask = (mask.dimx() > 0);
  int asize = 0;
  if (!hasmask)
    asize = d.dimx() * d.dimy();
  else {
    for (i = 0; i < mask.dimx(); ++i)
      for (j = 0; j < mask.dimy(); ++j)
        asize += (mask(i, j) > 0);
  }

  A.init(asize, 3);
  b.init(asize);
  int row = 0;
  for (i = xlow; i < d.dimx(); ++i)
    for (j = ylow; j < d.dimy(); ++j) {
      if (!hasmask || mask.get(i - xlow, j - ylow)) {
        A(row, 0) = i;
        A(row, 1) = j;
        A(row, 2) = 1;
        b(row) = d(i, j);
        ++row;
      }
    }

  calc_QR_decomp(A, sigma, gamma);
  calc_R(A, sigma, R);
  dgrid R1;
  R.subgrid(R1, 0, 0, R.dimy(), R.dimy());
  calc_Qtransb(A, gamma, b, c);
  dgrid c1;
  c.subgrid(c1, 0, 0, R.dimy(), 1);
  backward_solve(R1, c1, x);

  for (i = 0; i < d.dimx(); ++i)
    for (j = 0; j < d.dimy(); ++j)
      d(i, j) -= x(0) * i + x(1) * j + x(2);

} // dgrid::subtract_lsq

// __________________________________________________________________________
// Perform a median filter_ operation on an image 'times' times with
// a radius of 'size'

image& image::medfilt(int size, int times)
{
  int regsize = (2 * size + 1) * (2 * size + 1);
  ucgrid values(regsize);
  for (int count = 0; count < times; ++count)
  {
    image newimage(dimx(), dimy());
    newimage.clear();
    for (int i = 0; i < dimx() ; ++i)
      for (int j = 0; j < dimy() ; ++j)
      {
        int currval = 0;
        for (int k = -size; k <= size; ++k)
          for (int l = -size; l <= size; ++l)
	    values(currval++) = get(i + k, j + l);
	values.sort();
        newimage(i, j) = values(regsize / 2);
      }
    *this << newimage;
  }

  return *this;

} // image::medfilt

// __________________________________________________________________________
// Perform a mean filter operation on an image 'times' times with
// a radius of 'size'

image& image::meanfilt(int size, int times)
{
  int regsize = (2 * size + 1) * (2 * size + 1);
  ucgrid values(regsize);
  for (int count = 0; count < times; ++count)
  {
    image newimage(dimx(), dimy());
    for (int i = 0; i < dimx(); ++i)
      for (int j = 0; j < dimy(); ++j)
      {
        int currval = 0;
        for (int k = -size; k <= size; ++k)
          for (int l = -size; l <= size; ++l)
	    currval += (*this).get(i + k, j + l);
        newimage(i, j) = uchar(currval / regsize);
      }
    *this << newimage;
  }

  return *this;

} // image::meanfilt



image& image::histeq(int maxval)
{
  assert(maxval > 0 && maxval <= 255);

  int i, j, numvals = dimx() * dimy();
  igrid valcounts(maxval + 1), valeqs(maxval + 1);

  valcounts.clear();
  for (i = 0; i < dimx(); ++i)
    for (j = 0; j < dimy(); ++j)
      valcounts((*this)(i, j) * maxval / 255)++;

  for (i = 1; i < maxval + 1; ++i)
    valcounts(i) += valcounts(i - 1);

  valeqs(0) = ((1 + valcounts(0)) / 2) * maxval / numvals;
  for (i = 1; i < maxval + 1; ++i)
    valeqs(i) = (valcounts(i - 1) + 
		 (1 + valcounts(i) - valcounts(i - 1)) / 2) * maxval / numvals;

  for (i = 0; i < dimx(); ++i)
    for (j = 0; j < dimy(); ++j)
      (*this)(i, j) = uchar(valeqs((*this)(i, j)));

  return *this;

} // image::histeq

image& image::shift(int sx, int sy, int& x1, int& y1, int& x2, int& y2,
                     int& x3, int& y3)
{
  if (sx==0 && sy==0)
    return (*this);

  image newimage(dimx(),dimy());

  gridfun(get(i+sx, j+sy),newimage);
  x1 -= sx; x2-=sx; x3-=sx;
  y1 -= sy; y2-=sy; y3-=sy;

  (*this) << newimage;

  return (*this);  
}

// __________________________________________________________________________
// If 'amount' >= 0, put a border of 'amount' pixels on all edges of 'oldgrid'.
// If 'amount' < 0, center 'oldgrid' in an amount by amount grid.

image& image::imtrim(int amount, int& x1, int& y1, int& x2, int& y2,
                   int& x3, int& y3)
{
  int leftmar = 0, rightmar = 0, topmar = 0, bottommar = 0;

  int i, j, found = 0;
  for (i = 0; i < dimx() && !found; ++i)
    for (j = 0; j < dimy() && !found; ++j)
    {
      found = ((*this)(i, j) > 0);
      if (found)
        leftmar = i;
    }

  // Return now if the image is blank
  if (!found)
    return *this;

  found = 0;
  for (i = dimx() - 1; !found; --i)
    for (j = 0; j < dimy() && !found; ++j)
    {
      found = ((*this)(i, j) > 0);
      if (found)
        rightmar = i;
    }

  found = 0;
  for (j = 0; !found; ++j)
    for (i = 0; i < dimx() && !found; ++i)
    {
      found = ((*this)(i, j) > 0);
      if (found)
        topmar = j;
    }

  found = 0;
  for (j = dimy() - 1; !found; --j)
    for (i = 0; i < dimx() && !found; ++i)
    {
      found = ((*this)(i, j) > 0);
      if (found)
        bottommar = j;
    }

  int xlength = rightmar - leftmar + 1;
  int ylength = bottommar - topmar + 1;
  image newimage;
  if (amount >= 0)
  {
    x1 -= leftmar - amount;
    x2 -= leftmar - amount;
    x3 -= leftmar - amount;
    y1 -= topmar - amount;
    y2 -= topmar - amount;
    y3 -= topmar - amount;
    newimage.init(xlength + 2 * amount, ylength + 2 * amount);
    newimage.clear(BGD);
    for (i = amount; i < newimage.dimx() - amount; ++i)
      for (j = amount; j < newimage.dimy() - amount; ++j)
	newimage(i, j) = (*this)(i + leftmar - amount, j + topmar - amount);
  }
  else
  {
    amount *= -1;
    newimage.init(amount, amount);
    newimage.clear(BGD);

    int startx = ((amount > xlength) ? ((amount - xlength) / 2) : 0);
    int starty = ((amount > ylength) ? ((amount - ylength) / 2) : 0);
    x1 += startx; x2 += startx; x3 += startx;
    y1 += starty; y2 += starty; y3 += starty;
    for (i = 0; i < amount && i < xlength; ++i)
      for (j = 0; j < amount && j < ylength; ++j)
	newimage(i + startx, j + starty) = (*this)(i + leftmar, j + topmar);
  }
  //   for (i = 0; i < amount; ++i)
  //   for (j = 0; j < amount; ++j)
  //cout << int(newimage(i,j));

  (*this) << newimage;
  return (*this);

} // image::imtrim

// __________________________________________________________________________

image& image::make_spots(int num_spots, int spot_size)
{
  static int mask_made = 0;
  static igrid mask;

  // Make the mask if necessary.
  int i, j;
  if (!mask_made)
  {
    ++mask_made;
    mask.init(spot_size, spot_size);
    for (i = 0; i < spot_size; ++i)
      for (j = 0; j < spot_size; ++j) {
        mask(i, j) = int(1.0 / (pow(2.0, fabs(i - (spot_size - 1) / 2)) +
                                pow(2.0, fabs(j - (spot_size - 1) / 2))) * 
                         10000);
      }
  }

  image newimage = *this;
  for (i = 0; i < num_spots; ++i)
  {
    if (!randomize || rand() % 2 )  // 50% chance of a spot if randomizing.
    {
      int x = rand() % (dimx() - spot_size);
      int y = rand() % (dimy() - spot_size);
      for (j = 0; j < spot_size; ++j)
	for (int k = 0; k < spot_size; ++k)
	  newimage(x + j, y + k) = abs(newimage(x + j, y + k) - 
				       (rand() % 10000 < mask(j, k)));
    }
  }

//  (*this) << newimage.imtrim();

  return (*this);

} // image::make_spots

// __________________________________________________________________________

image& image::imblur(int blur, int& x1, int& y1, int& x2, int& y2,
		     int& x3, int& y3)
{
  if (blur <= 0)
    return *this;

  image newim(dimx() + 2 * blur, dimy() + 2 * blur);
  newim.clear();
  for (int i = 0; i < newim.dimx(); ++i)
    for (int j = 0; j < newim.dimy(); ++j)
    {
      int currpix = 0;
      for (int k = -blur; k <= blur; ++k)
	for (int l = -blur; l <= blur; ++l)
	  currpix += get(i + k - blur, j + l - blur);
      newim(i, j) = currpix / ((4 * blur) * (blur + 1) + 1);
    }
  *this << newim;

  x1 += blur; x2 += blur; x3 += blur; y1 += blur; y2 += blur; y3 += blur;

  return *this;

} // image::imblur

// __________________________________________________________________________

image& image::blur_downscale(double ratio, double sd)
{
  if (ratio == 1.0 && ( sd == 0 || datadepth > 1) )
    {
      normalize(255);
      return *this;
    }
  static int filter_made = 0;
  static dgrid filter;

  // Make the filter if necessary.
  if (!filter_made)
  {
   
    int i, j;
    // Calculate a truncated Gaussian convolution filter
    if (sd>0)
      {
	filter_made = 1;
	int center = int(floor(filter_size*sd / 2));
        cout << center << endl;
	filter.init(filter_size*sd + 1, filter_size*sd + 1);
	for (i = 0; i < filter.dimx(); ++i)
	  for (j = 0; j < filter.dimy(); ++j)
	    filter(i, j) = exp(-1 * ((i - center) * (i - center) + 
				     (j - center) * (j - center)) / (sd * sd));
	
	// Normalize the truncated filter so that it has norm 1.
	double sum = 0;
	for (i = 0; i < filter.dimx(); i++)
	  for (j = 0; j < filter.dimy(); j++)
	    sum += filter(i,j);
	for (i = 0; i < filter.dimx(); ++i)
	  for (j = 0; j < filter.dimx(); ++j)
	    filter(i, j) /= sum;
      }
     else
       {
        filter_size=1;
	filter.init(1,1);
        filter(0,0)=1;
       }
  }


  // Convolve and downscale the original image
  int newx = int(dimx() * ratio + FUZZ);
  int newy = int(dimy() * ratio + FUZZ);
  dgrid fnum(newx, newy);
  int fx2=filter.dimx()/2;
  int fy2=filter.dimy()/2;
  int i;
  for (i=0; i < newx; i++)
    for (int j=0; j < newy; j++)
    {
      int oldx = int(i / ratio + FUZZ);
      int oldy = int(j / ratio + FUZZ);
      double total=0;
      for (int k = 0; k < filter.dimx(); k++)
        for (int l = 0; l < filter.dimy(); l++)
         total += get(oldx - fx2 + k, oldy - fy2 + l) * filter(k, l);
      fnum(i, j) = total;
    }

  fnum.normalize(255);
  init(fnum.dimx(), fnum.dimy());
  for (i=0; i < dimx(); i++)
    for (int j=0; j < dimy(); j++)
      (*this)(i, j) = uchar(fnum(i, j));

  return (*this);

} // image::blur_downscale



image& image::downscale(int size)
{


  if (size <= 0 || dimx() == 0 || dimy() == 0)
    return *this;

  double rat = double(size)/double(MAX(dimx(), dimy()));

  image newimage(rint(dimx()*rat), rint(dimy()*rat));
 
  for (int i = 0; i < newimage.dimx(); ++i)
  {
    for (int j = 0; j < newimage.dimy(); ++j)
    {
	  newimage(i, j) = get( (int)floor(i/rat), (int)floor(j/rat));

    }
  }

  *this << newimage;

  return (*this);

} // downscale

// __________________________________________________________________________

image& image::imslant(int slnt)
{
  // Check to see we really want to slant.
  if (slnt <= 0)
    return *this;
  int imw=dimx();
  // Determine the average x and y indices of the stroke pixels.
  double yavg = 0.0, xavg = 0.0;
  int i, j, numstroke = 0;
  for (i = 0; i < dimx(); ++i)
    for (j = 0; j < dimy(); ++j)
      if ((*this)(i, j) != 0)
      {
        ++numstroke;
        xavg += i;
        yavg += j;
      }

  // Return now if the image is blank
  if (numstroke == 0)
    return *this;

  xavg /= numstroke;
  yavg /= numstroke;

  // Calculate the slope and intercept of the least squares line.
  double slopenum = 0.0, slopeden = 0.0;
  for (i = 0; i < dimx(); ++i)
    for (j = 0; j < dimy(); ++j)
      if ((*this)(i, j) != 0)
      {
        slopenum += j * (i - xavg);
        slopeden += j * (j - yavg);
      }
  double slope = (slopeden == 0.0 ? 0.0 : slopenum / slopeden);
  //  cout << slope << endl;
  //slope *= (fabs(slope) < 1.2);
//  double intercept = xavg - slope * yavg;
//
//   // Determine the MSE of the least squares line.
//   double mse = 0.0;
//   if (numstroke > 0)
//   {
//     for (i = 0; i < dimx(); ++i)
//       for (j = 0; j < dimy(); ++j)
//         if ((*this)(i, j) != 0)
//           mse += (slope * j + intercept - i) * (slope * j + intercept - i);
//     mse /= numstroke;
//   }

  // Correct the slant according to the slope calculated above.
  image newimage(dimx() * 2, dimy());
  newimage.clear();
  for (i = 0; i < dimx() * 2; ++i)
    for (j = 0; j < dimy(); ++j)
      {
	double fx= i - dimx() / 2 - (dimy() - j) * slope;
	double fy=j;
	int x = pint(fx);
	int y = pint(fy);
	double  a = fx - x;
	double  b = fy - y;
	int  xx1 = x + 1;
	int  yy1 = y + 1;   

	double z1 =  get(xx1,yy1,BGD) * a * b;
	double z2 =  get(x,yy1,BGD) * (1. - a) * b;
	double z3 =  get(xx1,y,BGD) * a * (1. - b);
	double z4 =  get(x,y,BGD) * (1. - a) * (1. - b);

      newimage(i, j) = uchar(z1+z2+z3+z4);
      }
  int x;
  *this << newimage.imtrim(-imw,x,x,x,x,x,x);
  
  return (*this);

} // image::imslant

// __________________________________________________________________________


// __________________________________________________________________________
// Regularly changing procedure...


image& image::std_process(double skw, double rtt, double slnt, double scl, int& x1, int& y1, int& x2, int& y2, int& x3, int& y3)
{

  
  if (randomize)
  {
    int randrtt = rand() % 1001;
    rtt *= double(randrtt - 500) / 500;

    if (skw > 0) 
      skw = exp((double)(5000 - rand() % 10000)/(10000*double(skw)));
    else 
      skw = 1;

    if (scl > 0)  
      scl = exp((double)(5000 -  rand() % 10000)/(10000*double(scl)));
    else 
      scl = 1;

    if (slnt > 0)  
      slnt = ((double)(5000 - rand() % 10000)/(10000*double(slnt)));
    else 
      slnt=0;
  }
  else
    {
      skw = exp(skw);
      scl = exp(scl);
      //      return(*this);
      //if (scl==0)
      //	scl=1;
      //if (skw==0)
      //	skw=1;
    }

  double cos_rotate = cos(rtt);
  double sin_rotate = sin(rtt);

  // Initialize the dimension of the newgrid sufficiently large.
  int rot_dim = 2 * MAX(dimx(), dimy());
  image newimage(rot_dim, rot_dim);
  double oldcenx = dimx() / 2;
  double oldceny = dimy() / 2;
  double newcenx = rot_dim / 2;
  double newceny = rot_dim / 2;
  for (int i = 0; i < rot_dim; ++i)
    for (int j = 0; j < rot_dim; ++j)
    {
      // Weird transform as images and sin/cos use different
      // coordinate systems.
      double dx = i - newcenx;
      double dy = j - newceny;
      double fx = oldcenx + scl*skw*(cos_rotate * dx - sin_rotate * dy);
      double fy = oldceny + scl/skw*(sin_rotate * dx + cos_rotate * dy);
      fx = fx + slnt*fy;
      int x = pint(fx);
      int y = pint(fy);
      double  a = fx - x;
      double  b = fy - y;
      int  xx1 = x + 1;
      int  yy1 = y + 1;   

      double z1 =  get(xx1,yy1,BGD) * a * b;
      double z2 =  get(x,yy1,BGD) * (1. - a) * b;
      double z3 =  get(xx1,y,BGD) * a * (1. - b);
      double z4 =  get(x,y,BGD) * (1. - a) * (1. - b);


      double rr = z1+z2+z3+z4;
      newimage(i, j) = uchar(rr);
    }

  double dx, dy;
  dx = (x1 - slnt*y1 - oldcenx)/(scl*skw);
  dy = (y1 - oldceny)*skw/scl;
  x1 = int(newcenx + cos_rotate * dx + sin_rotate * dy);
  y1 = int(newceny - sin_rotate * dx + cos_rotate * dy);
  dx = (x2 - slnt*y2 -oldcenx)/(scl*skw);
  dy = (y2 - oldceny)*skw/scl;
  x2 = int(newcenx + cos_rotate * dx + sin_rotate * dy);
  y2 = int(newceny - sin_rotate * dx + cos_rotate * dy);
  dx = (x3 -slnt*y3 - oldcenx)/(scl*skw);
  dy = (y3 - oldceny)*skw/scl;
  x3 = int(newcenx + cos_rotate * dx + sin_rotate * dy);
  y3 = int(newceny - sin_rotate * dx + cos_rotate * dy);

  (*this) << newimage.imtrim(0, x1, y1, x2, y2, x3, y3);

  return (*this);

} // image::std_process

// __________________________________________________________________________

// Read in the data for a digit.  'index' specified which digit to read in.
// Return before actually reading in the image if 'flag' is set.

int image::read(int index, int flag)
{
  // Simple error check
  if (index < 0)
    return 0;

  if (pgmfile!="")
   {
     int k=index;
     char buf[255];
     ifstream ifs;
     ifs.open(pgmfile);
      if (!ifs)
	{
	  cout << pgmfile << " doesn't exist " << endl;
	  exit(0);
	}
      for (int i=0; i<k; i++ )
	ifs.Getline(buf, 255);
      simpstrg comm;
      if(ifs.Getline(buf, 255))
	{
	  cout << buf <<endl;
	  simpstrg PGM=simpstrg("pgm");
	  //	  cout << " buf " << buf << endl;
	  if (strlen(buf)==0)
	    {
	      cerr << "Done\n";
	      exit(0);
	    }
	  if (strlen(buf)>0)// && !strstr(buf,PGM))
	    {
	      simpstrg nn=simpstrg(dataset)+simpstrg("/")+simpstrg(buf);
	      cout << " NN " << nn << endl;
	      //ifstream ifs(nn);
	      //if (!ifs.is_open())
	      //	{
	      //	  cerr << "Unable to read in non-pgm images\n";
	      //	  exit(0);
	      //	}
	      comm=simpstrg("convert ")+nn+" "+simpstrg("temp.pgm");
	      system(comm);
	      comm=simpstrg("temp.pgm");
	    }
	  else
	    comm=simpstrg(dataset)+simpstrg("/")+simpstrg(buf);
	}
      else
	{
	  cerr << " can't find file name " << buf << endl;
	  exit(0);
	}
     
      if (!loadpgm(comm))
      	return(0);
      return(1);
   }
  int digit_index = index % imagesperfile;
  int file_index = index / imagesperfile;
  simpstrg data_file = dataset + "/Images_" + ltos(file_index);


  // Open the data_file.
  ifstream ifs;
  if (!data_dir_open(ifs, data_file, 0))
      return 0;


  // Make sure we're getting the data for a digit that really exists.
  int entries;
  varread(ifs, entries);
  if (digit_index >= entries)
    return 0;

  // First skip to where the offset to the image data is, read in the offset,
  // and move to the offset.
  ifs.seekg((digit_index + 1) * Sint); 
  int offset;
  varread(ifs, offset);

  ifs.seekg(offset);

  // Read in the truth of the image.
  int truth;
  varread(ifs, truth);
//  tr = truth - '0';
  tr = truth;
  //if (flag>=0 && flag!=truth)
  // return(-1);
  // Read in the x-dimension and the y-dimension.
  int xdim, ydim;
  varread(ifs, xdim);
  varread(ifs, ydim);

  init(xdim, ydim);


  if (datadepth == 1)
  {
    int imsize = (xdim * ydim - 1) / 8 + 1;
    cvector imdata(imsize);
    ifs.read(imdata, imsize);
    for (int j = 0; j < ydim; ++j) {
      for (int i = 0; i < xdim; ++i) {
	int thebyte = (j * xdim + i) / 8;
	int thebit = 7 - (j * xdim + i) % 8;
	(*this)(i, j) = ((imdata[thebyte] >> thebit) & 1) * 255;
      } 
    }
  }
  else if (datadepth == 8) 
  {
    for (int j = 0; j < ydim; ++j)
      for (int i = 0; i < xdim; ++i)
	{
	  //	ifs.read(&((*this)(i, j)), 1);
	  uchar tmp;
	  varread(ifs,tmp); (*this)(i,j)=tmp;
	}
  }

  return 1;

} // image::read

/*
int image::read(int index, int flag)
{
  // Simple error check
  if (index < 0)
    return 0;

  // Initialize some parameters.
  int digit_index = index % imagesperfile;
  int file_index = index / imagesperfile;
  simpstrg data_file = dataset + "/Images_" + ltos(file_index);

  // Open the data_file.
  ifstream ifs;
  if (!data_dir_open(ifs, data_file, 0))
    {
      cerr << " Can not read image from file " << data_file << endl;
      exit(1);
    }
      


  // Make sure we're getting the data for a digit that really exists.
  int entries;
  varread(ifs, entries);
  if (digit_index >= entries)
    return 0;

  // First skip to where the offset to the image data is, read in the offset,
  // and move to the offset.
  ifs.seekg((digit_index + 1) * Sint); 
  int offset;
  varread(ifs, offset);
  ifs.seekg(offset);

  // Read in the truth of the image.
  int truth;
  varread(ifs, truth);
//  tr = truth - '0';
  tr = truth;

  // Read in the x-dimension and the y-dimension.
  int xdim, ydim;
  varread(ifs, xdim);
  varread(ifs, ydim);

  init(xdim, ydim);
  if (flag)
    return 1;

  if (datadepth == 1)
  {
    int imsize = (xdim * ydim - 1) / 8 + 1;
    cvector imdata(imsize);
    ifs.read(imdata, imsize);
    for (int j = 0; j < ydim; ++j) {
      for (int i = 0; i < xdim; ++i) {
	int thebyte = (j * xdim + i) / 8;
	int thebit = 7 - (j * xdim + i) % 8;
	(*this)(i, j) = ((imdata[thebyte] >> thebit) & 1) * 255;
      } 
    }
  }
  else if (datadepth == 8) 
  {
    for (int j = 0; j < ydim; ++j)
      for (int i = 0; i < xdim; ++i)
	ifs.read((char*)&((*this)(i, j)), 1);
  }

  return 1;

} // image::read

*/

// __________________________________________________________________________
// Option 94 Make a big latex display
int image::make_big(int sizex, int sizey, int maxent, int num,
		    int goodent, int numgood)
{
  double scl=scale;
  double sln=slant;
  randomize = 1;
  init(sizex, sizey);
  clear();

  cout << " num " << num << endl;
  cout << " goodent " << goodent << endl;
  cout << " numgood " << numgood << endl;

  vector<image> temp(1);
  temp[0].init(sizex, sizey);
  temp[0].clear();

  for (int i = 0; i < num + numgood; ++i)
  {
    int currind;
    if (i < num)
      do {
	currind = rand() % maxent;
      } while (currind == goodent);
    else 
      currind = goodent;
    //    cout << " i " << i << " " << currind << endl;
    int x = 0;
    image im;
    if(!im.read(currind+FIRSTDIM))
      {
	cerr << "Can't read image " << currind+FIRSTDIM <<" from "<< dataset << endl;
	exit(0);
      }
    im.imblur(1, x, x, x, x, x, x);
    if (synthesize)
      {
	void synth(image &im, int p10 = -1, int p11 = -1, 
           		      int p20 = -1, int p21 = -1,
		              int p30 = -1, int p31 = -1);

	image tim(64,64);
	tim.clear();
	int hd2 = MAX(64-MAX(im.dimx(),im.dimy()),0)/2;
	tim.subcopy(0,im.dimx(),0,im.dimy(),im,hd2,hd2);
	synth(tim, x, x, x, x, x, x);
	//vp.upd_grid();
	//vp.view();
	im=tim;
      }
    im.imtrim(-32, x, x, x, x, x, x);
    im.std_process(skew, rotate, sln, scl, x, x, x, x, x, x);
    //Get part of this image instead of all of it
    if (rand() % 1000 < dum2)
      {
	image tim(LBX,LBX);
        tim.clear();
        int xp = (rand() % im.dimx());
        int yp = (rand() % im.dimy());
	tim.subcopy(xp, xp+LBX, yp, yp+LBX,im,0,0);

        im=tim;
      }

    int topx = rand() % (sizex - im.dimx());
    int topy = rand() % (sizey - im.dimy());
    for (int j = 0; j < im.dimx(); ++j)
      for (int k = 0; k < im.dimy(); ++k)
	if (im(j, k) > dum1)
        {
	  set(topx + j, topy + k, im(j, k));
          if (currind==goodent)
	    temp[0].set(topx + j, topy + k, im(j, k));
	}
  }
  blur_downscale(1.,sd);
  write_images(temp, 1);
  system("mv Images_0 temp");
  return 1;

} // make_big

// __________________________________________________________________________

void view_image_list()
{
  Ifstream(ifs, listfile);
  if (!ifs)
  {
    cerr << "Unable to open listfile [" << listfile << "]\n";
    exit(1);
  }

  int indices[400], count = 0;
  while (ifs >> indices[count] && ++count < 400)
    ;

  pixsize = 1;
  int numperrow = MIN(20, int(sqrt(count) + .99));
  int isize = numperrow * ((int)scale + 2);

  image bigimage(isize, isize);
  bigimage.clear();
  for (int i = 0; i < count; ++i)
  {
    image tmpim;
    tmpim.read(indices[i]);
    tmpim.blur_downscale(ratio, sd);

    int starty = (i / numperrow) * ((int)scale + 2) + 1;
    int startx = (i % numperrow) * ((int)scale + 2) + 1;
    for (int ii = 0; ii < scale; ++ii)
      for (int j = 0; j < scale; ++j)
	bigimage(startx + ii, starty + j) = tmpim(ii, j);
  }
  view_normalized(bigimage);

} // view_image_list

// __________________________________________________________________________
// Given a digit name and index, returns the true value of the digit.
// Despite it's size, this routine should execute quite fast.

int get_truth(int digit)
{
  // Determine the name of the data_file and file index.
  int digit_index = digit % imagesperfile;
  int file_index = digit / imagesperfile;
  simpstrg data_file = dataset + "/Images_" + ltos(file_index);

  // Open the data_file.
  ifstream ifs;
  if (!data_dir_open(ifs, data_file))
    return -1;

  // Make sure we're getting the truth for a digit that really exists.
  int entries;
  varread(ifs, entries);
  if (digit_index >= entries)
    return -1;

  // First skip to where the offset to the image data is, read in the offset,
  // and move to the offset.
  ifs.seekg((digit_index + 1) * Sint); 
  int offset;
  varread(ifs, offset);
  ifs.seekg(offset);

  // Read in the truth of the image.
  int truth;
  varread(ifs, truth);

//  return truth - '0';
  return truth;

} // get_truth

// __________________________________________________________________________
// Try to open a file, first from the current directory, and then from the 
// directory 'dataname'.  Exit upon failure if 'abrt' flag is set
// Init global parameters if necessary.

int data_dir_open(ifstream& ifs, const simpstrg& dataname, int abrt)
{
  static int inited = 0;

  ifs.Open(dataname);
  if (ifs)
  {
    if (!inited)
    {
      inited = 1;
      Ifstream(ifs2, "Images_data");
      if (ifs2)
	ifs2 >> datadepth >> datasize;
    }
    return 1;
  }

  ifs.Open(datadir + "/" + dataname);
  if (!ifs && abrt)
  {
    cerr << "Unable to open [" << dataname << "] from data_dir_open" << endl;
    exit(1);
  }
  else if (ifs)
  {
    if (!inited)
    {
      inited = 1;
      Ifstream(ifs2, datadir + dataset + "/Images_data");
      if (ifs2)
	ifs2 >> datadepth >> datasize;
    }
    return 1;
  }

  return 0;

} // data_dir_open

// __________________________________________________________________________
// Randomly select 'num' integers from 0..'max'.  Put the values in 'l'
// If max > 250000 replication can occur.

void random_select(lgrid& randgrid, int max, int num)
{
  if (num <= 0 || max <= 0)
  {
    randgrid.init();
    return;
  }
  int nn=MIN(max,num);
  ucgrid tlist(max);
  tlist.clear();
  int cc=0;
  for (int k = 0; k < num && k < max; ++k)
  {

    if (num >= max)
      ++tlist[k];
    else if (num<max)
    {
      int tmp = lrand48() % max;
       while (tlist[tmp])
      {
        if (++tmp >= max)
	  tmp = 0;
      }
      ++tlist[tmp];
    }
    cc++;
  }

  randgrid.init(cc);
  cc=0;
  for (int k=0; k<max; k++)
    {
      if (tlist(k))
	{
        randgrid(cc)=k;
        cc++;
	}
    }    

  return;
  


} // random_select

// __________________________________________________________________________

void write_images(const vector<image>& im, int ndigits)
{
  for (int i = 0; i < ndigits; i += 100)
  {
    long k = i / 100;
    simpstrg ll = ltos(k);
    //cout << ll << endl;
    simpstrg imfile = simpstrg("Images_") + ll;
    //cout << imfile << endl;
    Ofstream(imf, imfile);

    int entries = MIN(100, ndigits - i);
    varwrite(imf, entries);

    int offset = Sint * (entries + 1);
    varwrite(imf, offset);
    int j;
    for (j = 1; j < entries; ++j)
    {
      offset += Sint + im[i + j - 1].ucgrid::size();
      varwrite(imf, offset);
    }
    for (j = 0; j < entries; ++j)
    {
      int imtruth = im[i+j].truth();
      varwrite(imf, imtruth);
      int dimx = im[i + j].dimx();
      int dimy = im[i + j].dimy();
      varwrite(imf, dimx);
      varwrite(imf, dimy);
      for (int l = 0; l < dimy; ++l)
        for (int k = 0; k < dimx; ++k)
          imf.write((char*)&im[i + j](k, l), 1);
    }
  }

} // write_images

// __________________________________________________________________________

void write_images(image* images, int numim, char* name)
{
  image bigim;
  int imagesize = MAX(images[0].dimx(), images[0].dimy());
  int bigsize = int(sqrt(numim) + 1) * (imagesize + 4);
  bigim.init(bigsize, bigsize);
  bigim.clear();
  int currx = 2, curry = 2;
  for (int i = 0; i < numim; ++i)
  {
    image writeim = images[i];
    writeim.normalize(0, 255);
    for (int j = 0; j < imagesize; ++j)
      for (int k = 0; k < imagesize; ++k)
        bigim(currx + j, curry + k) = writeim(j, k);
    currx += (imagesize + 4);
    if (currx + (imagesize + 4) > bigsize)
    {
      currx = 2;
      curry += (imagesize + 4);
    }
  }
  ucgrid& biggrid = bigim;
  biggrid.write(name);

} // write_images

// __________________________________________________________________________
// Convert a long to a string, return a pointer to the string.  This string
// is a global object and should be used with care.

char *ltos(const long x)
{
  static char tmpbuf[STRLEN];

  sprintf(tmpbuf, "%li", x);
  return tmpbuf;

} // ltos

double register_and_normalize(const image& im,  affinetrans& at, dgrid& regim,
                              ucgrid& pmask, int lsq)
{
  int i, j, k;

  double area = 0;
  double norm = 0;
  double normsquared = 0;
  int anymask=1;
  // In case the mask doesnt fit the image dont use it
  if (pmask.dimx()==0 || pmask.dimx()!=im.dimx() || pmask.dimy() != im.dimy())
    anymask=0;
  const int xsize = regim.dimx();
  const int ysize = regim.dimy();

  image tmpim(regim.dimx(), regim.dimy());
  
  norm = normsquared = 0;
  

  for (j = 0; j < xsize; ++j)
    for (k = 0; k < ysize; ++k)
      {

        if (anymask && !pmask(j, k)) 
	    tmpim(j, k) = 0;
	else
	  {
	    Pixel P = at * Pixel(j, k);
	    tmpim(j, k) = im.get(P[0], P[1]);
	    norm += tmpim(j, k);
	    normsquared += tmpim(j, k) * tmpim(j, k);
	    ++area;
	  }

      }
    

 
    
  double mean = norm / area;
  double deviation = sqrt(normsquared / area - mean * mean);
  
  if (deviation < 1) 
    return(0);
  
    for (i = 0; i < xsize; ++i)
      for (j = 0; j < ysize; ++j)
	regim(i, j) = (tmpim(i, j) - mean) / deviation;

    
  if (lsq)
    subtract_lsq(regim, pmask, 0, 0);

  return 1;

} // register_and_normalize

void register_to_dgrid(image& im, dgrid& regim, affinetrans at)
{
  for (int j = 0; j < regim.dimx(); ++j)
    for (int k = 0; k < regim.dimy(); ++k)
    {
      Pixel P = at * Pixel(j, k);
      if (P[0] >= 0 && P[0] < im.dimx() && P[1] >= 0 && P[1] < im.dimy())
	regim(j, k) = im(P[0], P[1]);
      else
	regim(j, k) = 100;
    }

} // register_to_dgrid

double l2_normalize(dgrid& d, const ucgrid& pmask)
{
  double area = 0;
  double norm = 0;
  double normsquared = 0;

  const int xsize = d.dimx();
  const int ysize = d.dimy();

  int i, j, k;

  if (normalstyle)
  {
    // REGION 1
    for (j = 0; j < xsize / 2; ++j)
      for (k = 0; k < ysize; ++k)
      {
        if (!pmask(j, k))
          continue;
	++area;
	norm += d(j, k);
	normsquared += d(j, k) * d(j, k);
      }

    double mean = norm / area;
    double deviation = sqrt(normsquared / area - mean * mean);

    for (i = 0; i < xsize / 2; ++i)
      for (j = 0; j < ysize; ++j)
	d(i, j) = (d(i, j) - mean) / deviation;

    // REGION 2
    area = norm = normsquared = 0;
    for (j = xsize / 2; j < xsize; ++j)
      for (k = 0; k < ysize; ++k) 
      {
        if (!pmask(j, k))
          continue;
	++area;
	norm += d(j, k);
	normsquared += d(j, k) * d(j, k);
      }

    mean = norm / area;
    deviation = sqrt(normsquared / area - mean * mean);

    for (i = xsize / 2; i < xsize; ++i)
      for (j = 0; j < ysize; ++j)
	d(i, j) = (d(i, j) - mean) / deviation;
  }
  else
  {
    for (j = 0; j < xsize / 2; ++j)
      for (k = 0; k < ysize; ++k)
      {
        if (!pmask(j, k))
          continue;
	++area;
	norm += d(j, k);
	normsquared += d(j, k) * d(j, k);
      }

    double mean = norm / area;
    double deviation = sqrt(normsquared / area - mean * mean);

    if (deviation < 1) 
      return(0);

    for (i = 0; i < d.dimx(); ++i)
      for (j = 0; j < d.dimy(); ++j)
	d(i, j) = (d(i, j) - mean) / deviation;

  }

  return 1;

} // l2_normalize




// __________________________________________________________________________

void make_procmask(ucgrid& processmask, int masktype, int sx, int sy)
{
  int i, j;

  processmask.init(sx, sy);

  switch (masktype) {
   case 0:
    processmask.clear(1);
    break;

   case 1: {
     processmask.clear();
     for (i = 0; i < sx; ++i)
       for (j = 0; j < sy; ++j)
         processmask(i, j) = 
           (i + j >= sy / 4 && i + j <= sy * 26 / 16 && 
            i - j <= sy * 11 / 16 && j - i <= sy * 11 / 16);
   }
   break;

   case 2: {
     processmask.clear();
     double cenx = (sx - 1) / 2.0;
     double ceny = (sy - 1) / 2.0;
     double distx, disty;
     for (i = 0; i < sx; ++i) {
       for (j = 0; j < sy; ++j) {
         distx = fabs(i - cenx);
         disty = fabs(j - ceny);
         if (distx == 0 && disty == 0)
           processmask(i, j) = 1;
         else {
           double cosang = distx / sqrt(distx * distx + disty * disty);
           double sinang = disty / sqrt(distx * distx + disty * disty);
           processmask(i, j) = 
             (fabs(i - cenx) < (sx / 2 - 2 * sx / 16) * cosang + .4 &&
              fabs(j - ceny) < sy / 2 * sinang + .4);
         }
       }
     }
   }
   break;

   case 3: {
     processmask.clear();
     double cenx = (sx - 1) / 2.0;
     double ceny = (sy - 1) / 2.0;
     double distx, disty;
     for (i = 0; i < sx; ++i) {
       for (j = 0; j < sy; ++j) {
         distx = fabs(i - cenx);
         disty = fabs(j - ceny);
         if (distx == 0 && disty == 0)
           processmask(i, j) = 1;
         else {
           double cosang = distx / sqrt(distx * distx + disty * disty);
           double sinang = disty / sqrt(distx * distx + disty * disty);
           processmask(i, j) = 
             (fabs(i - cenx) < (sx / 2 - 4 * sx / 16) * cosang + .8 &&
              fabs(j - ceny) < sy / 2 * sinang + .8);
         }
       }
     }
   }
   break;

   default:
    cerr << "Invalid mask type " << masktype << "in cart_best_question.\n";
    exit(1);
  }

} // make_procmask

void nistprocess(image &im)
{
  int x;	
  im.medfilt(1,1);
  im.imtrim(1,x,x,x,x,x,x);
  sd=1;
  if (scale==0)
    scale=32;
  if (threshold==0.)
    threshold=.2;
  filter_size=4;
  im.blur_downscale(1.,sd);
  int dd=MAX(im.dimy(),im.dimx());
  if (dd>scale)
    ratio=scale/dd;
  else 
    ratio=1.;
  if (ratio<1.)
    im.downscale((int)scale);
  im.normalize(255);
  if (threshold>=0)
    gridfun(im(i,j)> threshold*200, im);

  im.imtrim(-scale,x,x,x,x,x,x);
}

void get_subregions()
{

vector<images> imv(numltas);
for (int n=0; n<numltas; n++)
  {
    imv[n].init(numdata);
  }
igrid m;
igrid meanm(numltas,2);
meanm.clear();
int p=0;
int i=0;
 int done=0;
while (i<numdata && !done)
  {
    cout << " i " << i << endl;
     image im;
     if (im.read(first+i))
       {

	 im.normalize(imrange*numcolors,numcolors-1);
	 viewport vp(im);
	 vp.view();
	 vp.get_marked_points(m);
	 if (m.dimx()==numltas)
	   {
	     gridfun(m(i,j)+meanm(i,j),meanm);
	     for (int n=0; n<numltas; n++)
	       {
		 imv[n][p].init(2*regionsize, 2*regionsize);
		 imv[n][p].subcopy(m(n,0)-regionsize,m(n,0)+regionsize, 
				   m(n,1)-regionsize, m(n,1)+regionsize, im, 0, 0);
	       }
	     p++;
	 
	   }
	
       }
      else
        done=1;
     i++;
  }

 cout << " P="<<p<<endl;
ofstream ofs("meanlocs");

 for (int n=0; n<numltas; n++)
   {
     write_images(imv[n],p);
     ofs << meanm(n,0) << " ";
     ofs << meanm(n,1) << " " << endl;

     simpstrg r = simpstrg("mkdir ltaims")+ltos(n);
     system(r);
     simpstrg t = simpstrg("mv Images* ltaims")+ltos(n);
     system(t);
   }
}


int read_coords(igrid& point1, igrid& point2, igrid& point3, double scale, int ff, int ll)
{
  int i, j;
  int numd=ll-ff;
  point1.init(numd, 2);
  point2.init(numd, 2);
  point3.init(numd, 2);

  ifstream cfs;
  data_dir_open(cfs, dataset + "/coords", 0);
  if (!cfs)
  {
    point1.clear();
    point2.clear();
    point3.clear();
    return 0;
  }
    int tmp;
  for (i = 0; i < 6 * ff; ++i)
    cfs >> tmp;

  if (synthesize)
  {


    for (j = 0; j < 2; ++j)
    {
      cfs >> point1(0, j);
      point1(0, j) = int(point1(0, j) * scale + FUZZ);
    }
    for (j = 0; j < 2; ++j)
    {
      cfs >> point2(0, j);
      point2(0, j) = int(point2(0, j) * scale + FUZZ);
    }
    for (j = 0; j < 2; ++j)
    {
      cfs >> point3(0, j);
      point3(0, j) = int(point3(0, j) * scale + FUZZ);
    }
    for (i = 1; i < numd; ++i)
    {
      point1(i, 0) = point1(0, 0);
      point1(i, 1) = point1(0, 1);
      point2(i, 0) = point2(0, 0);
      point2(i, 1) = point2(0, 1);
      point3(i, 0) = point3(0, 0);
      point3(i, 1) = point3(0, 1);
    }
    return 1;
  }
 
  for (i = 0; i < numd; ++i)
  {
    for (j = 0; j < 2; ++j)
    {
      cfs >> point1(i, j);
      point1(i, j) = int(point1(i, j) * scale);
    }
    for (j = 0; j < 2; ++j)
    {
      cfs >> point2(i, j);
      point2(i, j) = int(point2(i, j) * scale);
    }
    for (j = 0; j < 2; ++j)
    {
      cfs >> point3(i, j);
      point3(i, j) = int(point3(i, j) * scale);
    }
  }

  return 1;

} // read_coords

// __________________________________________________________________________
// image.C

