// Copyright YALI AMIT, January 15, 2001.

// __________________________________________________________________________
// Subroutines to detect different types of local features.

#include "includes.h"
#include "landmark.h"
#include "image.h"
#include "pixel.h"
#include "globals.h"
#include "carttree.h"
#include "mask.h"

// Avector of Global 2d grids of image size used to store located features.
// A `1' at (i,j) means feature was found there. Index of grid indicates
// the feature type depending on the context.
extern ucgrids STAGS;



static int compar(const void *a, const void* b)
{
  return (*((const int*) a) < *((const int*) b));
}

void fill_STAGS(const landmarks& lm, int spread = 0, int numt = -1);
void empty_STAGS(const landmarks& lm);
void init_STAGS(int size, int dimx, int dimy);
void cluster_landmarks(lmarklist& lands, int clustsize, 
		       int numtypes, int& total);
void find_edge_in_block(const image& im, int x0, int y0, 
			       int& NbEdgesInTheBlock, igrid &TabEdges,
                               int unions);
void find_ridg_in_block(const image& im, int x0, int y0, 
			       int& NbEdgesInTheBlock, igrid &TabEdges,
                               int unions);
int find_mask(int i,int j, dgrid &dt,dgrid &mask, double *par, int hdim);
void cluster_landmarks(lmarklist& lands, int clustsize, 
		       int numtypes, int& total);

extern int codesize2;
// 8 nearest neighbors for detecting features for dynamic programming.
int SN[8][2] = { {0, 1}, {1, 1}, {1, 0}, {1,-1}, {0, -1}, {-1, -1}, {-1, 0}, {-1, 1}};
static int stagsflag = -1;
masks MS;

int Epsilon;

// Loop through blocks of size EdgeSizeBlock
// in the image, extract features and prune by keeping top
// EdgeNumberOfPixels of those in terms of contrast.
// Edgetype = 0 - discontinuity edges.
//           EdgeEpsilon - contrast lower bound.
//           EdgeGoodTests - number of inequalities out of the 6
//                           defined that need to be satisfied.
//           primitiveedges=0; contrast sensitive edges. 8 types.
//           primitiveedges=50; contrast insensitive edges. 4 types.
//           primitevedges=100; contrast sensitive, but no difference
//                              comparison. Simply difference > EdgeEpsilon.
//                              Same as EdgeGoodTests=-1 but computed faster.
//    EdgeType >= 1 Ridge detectors.
//            EdgeRadius - distancce to pixels for comparisons.
void detectedges(const image& im, landmarks& lmarks)
{

  // The minimal contrast threshold for an edge.
  Epsilon = int(rint(EdgeEpsilon*256));
  lmarks.set_truth(im.truth());
  if (EdgeType<3)
    lmarks.init_counts_offsets(8);
  // `ridg' detectors for curve tracking. All features detected
  //are used. No pruning.
  if (EdgeType>=1)
    {
      EdgeSizeBlock=1;
      EdgeNumberOfPixels=8;
    }
  int uni;
  // Contrast insensitive edges.
  uni = (primitiveedges==50);
  // Defaults: EdgeSizeBlock = 2
  int NbBlockPerLine = im.dimx() / EdgeSizeBlock;
  int NbBlockPerColumn = im.dimy() / EdgeSizeBlock;
  int numlands = 0;
  const int maxlands = 200000;
  static lmarklist land(maxlands);

  // Loop through disjoint blocks.
  igrid TabEdges(4, EdgeSizeBlock * EdgeSizeBlock * 8);
  for (int XBlock = 0; XBlock < NbBlockPerLine + 1; XBlock++)
  {
    for (int YBlock = 0; YBlock < NbBlockPerColumn + 1; YBlock++)
    {
      int x = XBlock * EdgeSizeBlock;
      int y = YBlock * EdgeSizeBlock;
      // for each edge TabEdges records 4 numbers
      // (0) the abs of the difference between the 2 central pixels
      // (1) the x coord of the edge
      // (2) the y coord of the edge
      // (3) the type : 0..7
      int NbEdgesInTheBlock = 0;
      // Find edges (discontinuity - local max of directional gradient).
      if (EdgeType==0)
	find_edge_in_block(im, x, y, NbEdgesInTheBlock, TabEdges, uni);
      // Ridges.
      else if (EdgeType==1 || EdgeType==2 )
	find_ridg_in_block(im, x, y, NbEdgesInTheBlock, TabEdges, uni);


      // Sort the found edges according to size of difference.
      if (NbEdgesInTheBlock > EdgeNumberOfPixels)
      {
	TabEdges.sort_columns(NbEdgesInTheBlock,compar);
      }

      // keep top EdgeNumberOfPixels
      int min = MIN(EdgeNumberOfPixels, NbEdgesInTheBlock);
      for (int i = 0; i < min && i + numlands < maxlands; i++)
      {
	land[i + numlands].cx = TabEdges(1, i);
	land[i + numlands].cy = TabEdges(2, i);
	land[i + numlands].index = TabEdges(3, i);
      }
      numlands += min;
    }
  }
  if (numlands > maxlands)
    numlands = maxlands;


  // Cluster edges
  cluster_landmarks(land, cluster, 8, numlands);
  // Store in a landmark structure.
  lmarks.setup_from_lmarklist(land, numlands);


} // Edge::DetectEdges

// __________________________________________________________________________
// Find the edges in the block by comparing absolute values.
void find_edge_in_block(const image& im, int x0, int y0, 
			       int& NbEdgesInTheBlock, igrid &TabEdges,
                               int unions)
{
  for (int x = 0; x < EdgeSizeBlock; x++)
  {
    for (int y = 0; y < EdgeSizeBlock; y++)
    {
      int j = x0 + x;
      int k = y0 + y;
      if (latexexp || ((j >= 2) && 
	              (j < im.dimx() - 2) && 
	              (k >= 2 ) && 
	              (k < im.dimy() - 2)))
      {
	Pixel T, C(j,k);
        double fac = 1;

	for (int i = 0; i < 4; i += edgeinc)
	{
	  switch (i)
	  {
	   case 0: {
             T = Pixel(0, -1);
             fac = 1;
             break;
           }
	   case 1: {
             T = Pixel(1, -1);
             fac = SQ2 / 2.0;
             break;
           }
	   case 2: {
             T = Pixel(1, 0);
             fac = 1;
             break;
           }
	   case 3: {
             T = Pixel(1, 1);
             fac = SQ2 / 2.0;
             break;
           }
	  }

	  Pixel R = T.Rot90();
	  Pixel P1 = C + T;

	  int v1 = im.get(P1[0], P1[1]);
          int vc = im.get(C[0], C[1]);

	  int diff = int(abs(vc - v1) * fac);
	  if (diff > Epsilon) 
	  {
	    Pixel P2 = P1 + R;
	    Pixel P3 = P1 + T;
	    Pixel P4 = P1 - R;
	    Pixel P5 = C + R;
	    Pixel P6 = C - T;
	    Pixel P7 = C - R;

	    int v2 = im.get(P2[0],P2[1]);
	    int v3 = im.get(P3[0],P3[1]);
	    int v4 = im.get(P4[0],P4[1]);
	    int v5 = im.get(P5[0],P5[1]);
	    int v6 = im.get(P6[0],P6[1]);
	    int v7 = im.get(P7[0],P7[1]);

	    if (primitiveedges==100 ||
		((diff > abs(v1-v3)) + (diff > abs(v1-v2)) +
		 (diff > abs(v1-v4)) +(diff > abs(vc-v5)) +
		 (diff > abs(vc-v6)) + (diff > abs(vc-v7)) ) > EdgeGoodTests )
                 
	    {
              if (unions) {
		TabEdges(0, NbEdgesInTheBlock) = diff;
		TabEdges(1, NbEdgesInTheBlock) = j;
		TabEdges(2, NbEdgesInTheBlock) = k;
		TabEdges(3, NbEdgesInTheBlock) = 2*i;
		NbEdgesInTheBlock++;
              }
              else if (vc > v1)
	      {
		TabEdges(0, NbEdgesInTheBlock) = diff;
		TabEdges(1, NbEdgesInTheBlock) = j;
		TabEdges(2, NbEdgesInTheBlock) = k;
		TabEdges(3, NbEdgesInTheBlock) = 2 * i;
		NbEdgesInTheBlock++;
	      }
	      else
	      {
		TabEdges(0, NbEdgesInTheBlock) = diff;
		TabEdges(1, NbEdgesInTheBlock) = j;
		TabEdges(2, NbEdgesInTheBlock) = k;
		TabEdges(3, NbEdgesInTheBlock) = 2 * i + 1;
		NbEdgesInTheBlock++;
	      }
	    }
	  }
	}
      }
    }
  }

} // Edge::FindEdgesInTheBlock

// Find the ridg in the block.
void find_ridg_in_block(const image& im, int x, int y, 
			       int& NbEdgesInTheBlock, igrid &TabEdges,
                               int unions)
{
  
  int ln = 8;
  for (int k=0; k<ln; k++)
    {
      int ll = ( (k+2) % 8);
      int nx = EdgeRadius*SN[k][0];
      int ny = EdgeRadius*SN[k][1];
      int nlx = EdgeRadius/2*SN[ll][0];
      int nly = EdgeRadius/2*SN[ll][1];
      int logic=0;
      if (EdgeType==1)
	{
	  logic = (im.get(x,y) > im.get(x+nx,y+ny) &&
		   im.get(x,y) > im.get(x-nx,y-ny) &&
		   fabs(im.get(x,y)-im.get(x+nlx,y+nly))
		   < fabs(im.get(x,y)-im.get(x+nx,y+ny)) &&
		   fabs(im.get(x,y)-im.get(x+nlx,y+nly)) 
		   < fabs(im.get(x,y)-im.get(x-nx,y-ny)) );
		  
	  }
        else if (EdgeType==2)
	  {
	    logic = (im.get(x,y) < im.get(x+nx,y+ny) &&
		   im.get(x,y) < im.get(x-nx,y-ny) &&
		   fabs(im.get(x,y)-im.get(x+nlx,y+nly))
		   < fabs(im(x,y)-im.get(x+nx,y+ny)) &&
		   fabs(im.get(x,y)-im.get(x+nlx,y+nly)) 
		   < fabs(im.get(x,y)-im.get(x-nx,y-ny)) );
	  }
	

      if (logic)
	{
	  TabEdges(0, NbEdgesInTheBlock) = 0;
	  TabEdges(1, NbEdgesInTheBlock) = x;
	  TabEdges(2, NbEdgesInTheBlock) = y;
	  TabEdges(3, NbEdgesInTheBlock) = k;
	  NbEdgesInTheBlock++;	
	}
    }
  
}

// Hard wired pairs of edges. There are 8 neighborhoods corresponding
// to 8 nearest neighbors of center edge. Each neighborhood is a square
// radius `spreadforpair', and is centered EdgeInterPixelDistance from
// center edge (sup-norm) distance. CLustering here is done by direct
// subsampling of the grid by `ltacluster' for recording the locations
// of the pairs.
void detectpairs(landmarks& edges, landmarks& ltalands)
{
  const int maxlands = 200000;
  static lmarklist land(maxlands);

  igrid sn(8,2);
  
  gridfun(SN[i][j]*EdgeInterPixelDistance, sn);
  // Spread the detected edges in neighborhoods of radius spreadforpair.
  // Makes finding the pairwise conjunctions more efficient.
  int numlt=256;
  fill_STAGS(edges, spreadforpair,numlt);

  ltalands.set_truth(edges.truth());
  for (int i = 0; i < 6; ++i)
    ltalands.coords[i] = edges.coords[i];
  ltalands.init_counts_offsets(numlt);
  int numlands=0;
  for (int k=0; k<8; k++)
    {
      int numt=edges.type_count(k);
      for (int i=0; i<numt; i++)
	{
	  int x=edges.location_x(k, i);
	  int y=edges.location_y(k, i);
	  for (int l=0; l<8; l++)
	    {
	      for (int n=0; n<4; n++)
		{
		  int index=n*64+8*l+k;
		  // These are coaligned pairs, too common to use.
		  if (index==0 || index==9 || index==164 || index==173)
		    continue;

		  if (STAGS[l].get(x+sn(n,0),y+sn(n,1)))
		    {
		      if (coarse)
			{
			  land[numlands].cx = x/ltacluster;
			  land[numlands].cy = y/ltacluster;
			  land[numlands].index = index;
			}
		      else
			{
			  land[numlands].cx = x;
			  land[numlands].cy = y;
			  land[numlands].index = index;
			}
			  numlands++;
			
		      if (numlands == maxlands)
			exit(0);
		    }
		}
	    }
	}
    } 
  empty_STAGS(edges);
  
  if (!coarse)
    cluster_landmarks(land, cluster, numlt, numlands);
  // Store remaining ltas in a landmark structure
  ltalands.setup_from_lmarklist(land, numlands);

}



static int com(const void *a, const void* b)
{
  return (*((const double* )(a)) > *((const double*)(b)));
}


// Detect masks of comparisons as used in graph matching algorithm.
// EdgeEpsilon is again minimum contrast for a difference to be counted
// as positive or negative. The masks, the thresholds for the positive
//and negative differences are given in the file `masks.asc' in directory
//`maskdir'
void detectmasks(image &im, landmarks &ltalands)
{

  const int maxlands = 200000;
  lmarklist land(maxlands);

  if (MS.dim()==0)
    {

      simpstrg filn=maskdir+simpstrg("/masks.asc");
      ifstream ifs(filn);
      int num;
      ifs >> num;
      MS.init(num);
     
      for (int m=0; m < MS.dim(); m++)
	MS[m].read(ifs);
      EdgeEpsilon *= 255;
    }  

  dgrid datat(1,im.dimx()*im.dimy());
  int p=0;
  for (int i=0; i<im.dimx(); i++)
    for (int j=0; j<im.dimy(); j++)
      {
	datat(p)=double(im(i,j));
	p++;
      }
  datat.sort_columns(datat.dimy(),com);
  
  int numls=0;
  
  
  for (int m=0; m < MS.dim(); m++)
    {

      double fracmx = datat[(int)((MS[m].frac*.9+.05)*datat.dimy())];
      //      cout << "fracmx " << fracmx << endl;
      // Find points satisfying criterion 
      for (int i=0;i<im.dimx();i+=1)
	for (int j=0;j<im.dimy();j+=1)
	  {
	    if ( im(i,j) > fracmx*threshold && MS[m].find(i,j,im) )
	      {
		land[numls].cx = i;
		land[numls].cy = j;
		land[numls].index = m;
		numls++;
	      } 
	  }
    }

  cluster_landmarks(land, cluster,  MS.dim(), numls);
  ltalands.init_counts_offsets(MS.dim());
  ltalands.setup_from_lmarklist(land, numls);
}


int mask::find(int i, int j, image &im)
{

 int ca = 0, cb = 0;
 for (int k=0;k<msk.dimx();k++)
   for (int l=0;l<msk.dimx();l++)
     {
       if ( (i != k || j != l) && msk(k,l) != 0    )
         {
	   int diff = im(i,j) - im.get(i-hdim+k,j-hdim+l);
	   if ( msk(k,l)==-1 && diff<=EdgeEpsilon) cb+=1;
	   if ( msk(k,l)== 1 && diff>=EdgeEpsilon) ca+=1;
	 }
     }  
 return (ca>=par[0] && cb >= par[1]); 
}



// Detect the tags in terms of a classification
// tree classifying `codesize' by `codesize'
// binary microimages.
void detecttags(image &im, landmarks &ltalands, carttree &tagtree)
{
  int check_bgd(ucgrid &in);
  ltalands.init_counts_offsets(numltas);
  ltalands.set_truth(im.truth());
  dgrid cartim(codesize, codesize);
  ucgrids uc;
  
  int numlands=0;
  static lmarklist land(maxlands);
  for (int k=-codesize+1; k<im.dimx(); k++)
    for (int l=-codesize+1; l<im.dimy(); l++)
      {
	ucgrid uccart(codesize,codesize);
	uccart.subcopy(k, k+codesize, l, l+codesize, im, 0, 0);
	
	if ( !check_bgd(uccart))
	  continue;
	else
	  {
	    gridfun(uccart(i,j), cartim);
	  }
	cartiter terminal;
	tagtree.dropdata(cartim, uc, terminal);
	//     cout << terminal.history() << endl;
	int g[terminal.depth()];
	int n=terminal.depth()-1;
	cartiter term=terminal;
	while (term.parent())
	  {
	    g[n]=term.yn();
	    term=term.parent();
	    n=n-1;
	  }
	int ind=-1;
	for (int m=0; m<terminal.depth(); m++)
	  {
	    
	    land[numlands].cx = k+codesize2;
	    land[numlands].cy = l+codesize2;
	    
	    if (g[m])
	      ind = ind *2 + 3;
	    else
	      ind = ind *2 + 2;
	    land[numlands].index = ind;
	    //cout << land[numlands].cx << " " << land[numlands].cy << " " << ind << endl;
	    numlands++;
	  }
	
	
      }
  // cout << " NUM" << numlands << endl;
  cluster_landmarks(land, ltacluster, numltas, numlands);
  ltalands.setup_from_lmarklist(land, numlands);
  
}

// Check if subimage is mostly constant, then it is not interesting.        
int check_bgd(ucgrid &in)
{
  int stroke = 0;
  int low=0,high=3;
  //if (codesize==4)
  //  low =1;
  low=codesize/2-1;
  high=codesize/2+1;
  /*
  else if (codesize<=3)
    { 
      low = 0;
      if (codesize==2) high=2;
    } 
  */
  
  for (int l = low; l < high; ++l)
    {
      for (int m = low; m < high; ++m)
	{
	  stroke += (in.get(l, m) > 0);
	}
    }
  if (stroke == 0 || stroke == (high-low) * (high-low))
    return 0;
  else 
    return 1;
  
  
}


// FIll the global STAGS buffers with some spreading in boxes of radius
// spread
void fill_STAGS(const landmarks& lm, int spread, int numt)
{

  if (stagsflag != -1)
  {
    cerr << "Filling an already filled STAGS\n";
    exit(1);
  }
  int numta;
  int aspread=( (spread >= 0) ? (spread) : (-spread) );
  stagsflag = aspread;
  static int inited = 0;

  // Needs to be initialized only once
  if (!inited)
  {
    //    cout << "INITING STAGS " << numt << endl;
    if (numt<0)
       numta=lm.total_types();
    else
        numta=numt;
    inited=1;
    int nn=MAX(STAGSsize, numta);
    init_STAGS(nn, STAGSdim, STAGSdim);
  }

  numta = lm.total_types();
  for (int currtype = 0; currtype < lm.total_types(); currtype++)
  {
    const int numins = lm.type_count(currtype);
    for (int currinst = 0; currinst < numins; ++currinst)
    {
      const int x = lm.location_x(currtype, currinst);
      const int y = lm.location_y(currtype, currinst);
      if (spread != 0)
      {
	const int minx = MIN(aspread, x);
	const int miny = MIN(aspread, y);
	const int maxx = MIN(aspread, STAGS[0].dimx() - x - 1);
	const int maxy = MIN(aspread, STAGS[0].dimy() - y - 1);
	if (spread > 0)
	  {
	    for (int i = -minx; i <= maxx; ++i)
	      for (int j = -miny; j <= maxy; ++j) 
		{
		  if (STAGS[currtype](x + i, y + j) == 0)
		    STAGS[currtype](x + i, y + j) = 2;
		}
	  }
	// Spreading along lines not boxes
        else if (spread<0)
	  {
	    switch (currtype)
	      {
	      case 0:
	      case 1:
		{
		  for (int j = -miny; j <= maxy; ++j) 
		    {
		      if (STAGS[currtype](x, y + j) == 0)
			STAGS[currtype](x, y + j) = 2;
		    }
		}
		break;
	      case 2:
	      case 3:
		{
		  for (int j = -miny; j <= maxy; ++j) 
		    {
 		      if (-j>=-minx && -j<=maxx)
 			{
			  if (STAGS[currtype](x-j, y + j) == 0)
			    STAGS[currtype](x-j, y + j) = 2;
			}
		    }
		}
		break;
	      case 4:
              case 5:
		{
		  for (int i = -minx; i <= maxx; ++i)
		    {
		      if (STAGS[currtype](x + i,y) == 0)
			STAGS[currtype](x + i,y) = 2;
		    }
		}
		break;
	      case 6:
	      case 7:
		{
		  for (int j = -miny; j <= maxy; ++j) 
		    {
		    if (j>=-minx && j<=maxx)
		      {
			if (STAGS[currtype](x+j, y + j) == 0)
			  STAGS[currtype](x+j, y + j) = 2;
		      }
		    }
		}
		break;
	      }
	  }
      }
          
      STAGS[currtype](x, y) = 1;
    }
  }

} // fill_STAGS


// EMpty global Stags buffers
void empty_STAGS(const landmarks& lm)
{
  if (stagsflag < 0)
  {
    cerr << "Emptying an already empty STAGS\n";
    exit(1);
  }

  const int numt = lm.total_types();
  for (int currtype = 0; currtype < numt; currtype++)
  {
    const int numins = lm.type_count(currtype);
    for (int currinst = 0; currinst < numins; ++currinst)
    {
      const int x = lm.location_x(currtype, currinst);
      const int y = lm.location_y(currtype, currinst);
      if (stagsflag != 0)
      {
	const int minx = MIN(stagsflag, x);
	const int miny = MIN(stagsflag, y);
	const int maxx = MIN(stagsflag, STAGS[0].dimx() - x - 1);
	const int maxy = MIN(stagsflag, STAGS[0].dimy() - y - 1);
        for (int i = -minx; i <= maxx; ++i)
          for (int j = -miny; j <= maxy; ++j)
	    STAGS[currtype](x + i, y + j) = 0;
      }
      STAGS[currtype](x, y) = 0;
    }
  }

  stagsflag = -1;

} // empty_STAGS

// __________________________________________________________________________
// Initialize global STAGS buffers to code locations of edges or ltas
void init_STAGS(int size, int dimx, int dimy)
{
  for (int i = 0; i < size; ++i) {
    STAGS[i].init(dimx, dimy);
    STAGS[i].clear();
  }

} // init_STAGS

